// SPDX-FileCopyrightText: 2021-2025 The Ikarus Developers ikarus@ibb.uni-stuttgart.de
// SPDX-License-Identifier: LGPL-3.0-or-later

#pragma once
#include <ikarus/utils/makeenum.hh>

namespace Ikarus {
MAKE_ENUM(DBCOption, Raw, Reduced, Full);
}